
--[[
	LibFeedback is under development ATM
	It works as a nice replacement for the message frame in my addons so far
	Eventually it will have more animations, use StickyFrames and handle BigWigs-like alerts
	Some code ideas were borrowed, and probably badly-reused, from LibSharedMedia and Parrot
]]

local MAJOR_VERSION = "LibFeedback-0.1"
local MINOR_VERSION = tonumber(("$Revision: 65547 $"):match("%d+"))

local lib = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

local _G = _G
local exp = _G.exp
local type = _G.type
local pairs = _G.pairs
local tinsert = _G.tinsert
local tremove = _G.tremove
local random = _G.random
local strsplit = _G.strsplit
local strupper = _G.strupper
local UIParent = _G.UIParent
local PlaySoundFile = _G.PlaySoundFile
local CallbackHandler = LibStub:GetLibrary("CallbackHandler-1.0")

local FEEDBACK_ICON = "questionmark.ico"
local FEEDBACK_FONTCOLOR = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 }
local FEEDBACK_FONTTYPE = "Fonts\\ARIALN.TTF"
local FEEDBACK_FONTSIZE = 16
local FEEDBACK_FONTOUTLINE = ""
local FEEDBACK_ANIMATION = "Scroll"
local FEEDBACK_DIRECTION = "UP"
local FEEDBACK_DELAY = 0.025
local FEEDBACK_STEPS = 64
local FEEDBACK_STICK = 16
local FEEDBACK_ANIMATIONS = {
	["Scroll"] = {
		desc = "Move messages in a sliding motion",
		create = function(feedbackFrame, feedbackObject)
			local frame, object, previousobject = lib.frames[feedbackFrame], lib.objects[feedbackFrame][feedbackObject], lib.objects[feedbackFrame][feedbackObject - 1]
			local direction1, direction2 = strsplit(":", strupper(frame.direction))
			lib.objects[feedbackFrame][feedbackObject].step = object.step + 1
			if previousobject then -- fix vertical overlapping - may need some work
				if direction1 == "DOWN" and object.yoffset > previousobject.yoffset + previousobject.fontsize then
					lib.objects[feedbackFrame][feedbackObject].yoffset = previousobject.yoffset + previousobject.fontsize
				elseif direction1 == "UP" and object.yoffset > previousobject.yoffset - previousobject.fontsize then
					lib.objects[feedbackFrame][feedbackObject].yoffset = previousobject.yoffset - previousobject.fontsize
				elseif direction2 == "ALT" and (direction1 == "DOWN" or direction1 == "UP") and exp(object.number) ~= 0 and object.yoffset > previousobject.yoffset + previousobject.fontsize then
				--	lib.objects[feedbackFrame][feedbackObject].xoffset = previousobject.xoffset + previousobject.fontsize
					lib.objects[feedbackFrame][feedbackObject].yoffset = previousobject.yoffset + previousobject.fontsize
				elseif direction2 == "ALT" and (direction1 == "DOWN" or direction1 == "UP") and exp(object.number) == 0 and object.yoffset > previousobject.yoffset - previousobject.fontsize then
				--	lib.objects[feedbackFrame][feedbackObject].xoffset = previousobject.xoffset - previousobject.fontsize
					lib.objects[feedbackFrame][feedbackObject].yoffset = previousobject.yoffset - previousobject.fontsize
				end
			end
			lib.objects[feedbackFrame][feedbackObject].fontstring = lib.frames[feedbackFrame]:CreateFontString("Feedback_" .. feedbackFrame .. object.number .. "_FontString", "ARTWORK", "MasterFont")
			lib.objects[feedbackFrame][feedbackObject].fontstring:SetPoint("CENTER", frame, "CENTER", object.xoffset, object.yoffset)
			lib.objects[feedbackFrame][feedbackObject].fontstring:SetFont(object.fonttype, object.fontsize, object.fontoutline)
			lib.objects[feedbackFrame][feedbackObject].fontstring:SetTextColor(object['fontcolor'].r, object['fontcolor'].g, object['fontcolor'].b)
			lib.objects[feedbackFrame][feedbackObject].fontstring:SetText(object.text)
			lib.objects[feedbackFrame][feedbackObject].fontstring:SetParent(frame)
			lib.objects[feedbackFrame][feedbackObject].fontstring:Show()
			if object.sound then
				PlaySoundFile(object.sound)
			end
			if object.visual then
				object.visual()
			end
		end,
		modify = function(feedbackFrame, feedbackObject)
			local frame, object, previousobject = lib.frames[feedbackFrame], lib.objects[feedbackFrame][feedbackObject], lib.objects[feedbackFrame][feedbackObject - 1]
			local direction1, direction2 = strsplit(":", strupper(frame.direction))
			lib.objects[feedbackFrame][feedbackObject].step = object.step + 1
			if direction1 == "LEFT" then
				lib.objects[feedbackFrame][feedbackObject].xoffset = object.xoffset - 4
			elseif direction1 == "RIGHT" then
				lib.objects[feedbackFrame][feedbackObject].xoffset = object.xoffset + 4
			elseif direction1 == "DOWN" then
				lib.objects[feedbackFrame][feedbackObject].yoffset = object.yoffset - 3
			elseif direction1 == "UP" then
				lib.objects[feedbackFrame][feedbackObject].yoffset = object.yoffset + 3
			end
			if direction2 == "LEFT" then
				lib.objects[feedbackFrame][feedbackObject].xoffset = object.xoffset - 4
			elseif direction2 == "RIGHT" then
				lib.objects[feedbackFrame][feedbackObject].xoffset = object.xoffset + 4
			elseif direction2 == "DOWN" then
				lib.objects[feedbackFrame][feedbackObject].yoffset = object.yoffset - 3
			elseif direction2 == "UP" then
				lib.objects[feedbackFrame][feedbackObject].yoffset = object.yoffset + 3
			elseif direction2 == "ALT" and (direction1 == "DOWN" or direction1 == "UP") and exp(object.number) ~= 0 then
				lib.objects[feedbackFrame][feedbackObject].xoffset = object.xoffset - 4
			elseif direction2 == "ALT" and (direction1 == "DOWN" or direction1 == "UP") and exp(object.number) == 0 then
				lib.objects[feedbackFrame][feedbackObject].xoffset = object.xoffset + 4
			elseif direction2 == "ALT" and (direction1 == "LEFT" or direction1 == "RIGHT") and exp(object.number) ~= 0 then
				lib.objects[feedbackFrame][feedbackObject].yoffset = object.yoffset - 3
			elseif direction2 == "ALT" and (direction1 == "LEFT" or direction1 == "RIGHT") and exp(object.number) == 0 then
				lib.objects[feedbackFrame][feedbackObject].yoffset = object.yoffset + 3
			end
			if object.sticky and object.step > frame.steps - 10 then
				local shakex, shakey = random(-4, 4), random(-3, 3)
				lib.objects[feedbackFrame][feedbackObject].fontstring:SetPoint("CENTER", frame, "CENTER", object.stickx + shakex, object.sticky + shakey)
			elseif object.sticky and object.step == frame.steps - 10 then 
				lib.objects[feedbackFrame][feedbackObject].fontstring:SetFont(object.fonttype, object.fontsize + 8, object.fontoutline)
				lib.objects[feedbackFrame][feedbackObject].fontstring:SetPoint("CENTER", frame, "CENTER", object.xoffset, object.yoffset)
				_, _, _, object.stickx, object.sticky = lib.objects[feedbackFrame][feedbackObject].fontstring:GetPoint()
			elseif object.step < frame.steps and object.step > frame.steps - 10 then
				lib.objects[feedbackFrame][feedbackObject].fontstring:SetAlpha((frame.steps - object.step) / 10)
				lib.objects[feedbackFrame][feedbackObject].fontstring:SetPoint("CENTER", frame, "CENTER", object.xoffset, object.yoffset)
			elseif object.step < frame.steps then
				lib.objects[feedbackFrame][feedbackObject].fontstring:SetPoint("CENTER", frame, "CENTER", object.xoffset, object.yoffset)
			end
		end,
		destroy = function(feedbackFrame, feedbackObject)
			lib.objects[feedbackFrame][feedbackObject].fontstring:Hide()
			tremove(lib.objects[feedbackFrame], feedbackObject)
		end,
	}
}

lib.frames = lib.frames or {}
lib.objects = lib.objects or {}
lib.callbacks = lib.callbacks or CallbackHandler:New(lib)
lib.mainframe = lib.mainframe or CreateFrame("Frame", "Feedback", UIParent)
lib.mainframe:SetScript("OnUpdate", function(this, elapsed)
	for feedbackFrame in pairs(lib.frames) do
		local frame = lib.frames[feedbackFrame]
		if frame and lib.objects[feedbackFrame] then
			for feedbackObject in pairs(lib.objects[feedbackFrame]) do
				local object = lib.objects[feedbackFrame][feedbackObject]
				if object and object.step then
					if object.step == 0 then
						lib.objects[feedbackFrame][feedbackObject].elapsed = 0
						FEEDBACK_ANIMATIONS[frame.animation].create(feedbackFrame, feedbackObject)
					elseif object.step < frame.steps or (object.sticky and (object.step < frame.steps + frame.stick)) then
						lib.objects[feedbackFrame][feedbackObject].elapsed = object.elapsed + elapsed
						if object.elapsed > frame.delay then
							lib.objects[feedbackFrame][feedbackObject].elapsed = 0
							FEEDBACK_ANIMATIONS[frame.animation].modify(feedbackFrame, feedbackObject)
						end
					else
						FEEDBACK_ANIMATIONS[frame.animation].destroy(feedbackFrame, feedbackObject)
					end
				end
			end
		end
	end
end)

--[[
Notes:
	Create a feedback frame
Arguments:
	string - the frame name
	table [optional] - the frame attributes
		table.parent string - the frame parent [default = "UIParent"] 
		table.strata string - the frame strata [default = "HIGH"]
		table.width number - the frame width [default = 1]
		table.height number - the frame height [default = 1]
		table.xoffset number - the frame relative xoffset [default = 0]
		table.yoffset number - the frame relative yoffset [default = 0]
		table.animation - the type of animation to associate with the message [default = "Slide"]
		table.direction - the direction[s] to associate with the message animation [default = "UP"]
Example:
	local FB = LibStub("LibFeedback-0.1")
	FB:Create("MyFeedback")
]]
function lib:Create(feedbackFrame, feedbackAttributes)
	if feedbackFrame and type(feedbackFrame) == 'string' and not lib.frames[feedbackFrame] then
		lib.frames[feedbackFrame] = CreateFrame("Frame", "Feedback_" .. feedbackFrame, lib.mainframe)
		lib.frames[feedbackFrame].parent = feedbackAttributes and feedbackAttributes.parent or UIParent
		lib.frames[feedbackFrame].strata = feedbackAttributes and feedbackAttributes.strata or "HIGH"
		lib.frames[feedbackFrame].width = feedbackAttributes and feedbackAttributes.width or 128
		lib.frames[feedbackFrame].height = feedbackAttributes and feedbackAttributes.height or 64
		lib.frames[feedbackFrame].xoffset = feedbackAttributes and feedbackAttributes.xoffset or 0
		lib.frames[feedbackFrame].yoffset = feedbackAttributes and feedbackAttributes.yoffset or 0
		lib.frames[feedbackFrame].fontcolor = feedbackAttributes and feedbackAttributes.fontcolor or FEEDBACK_FONTCOLOR
		lib.frames[feedbackFrame].fonttype = feedbackAttributes and feedbackAttributes.fonttype or FEEDBACK_FONTTYPE
		lib.frames[feedbackFrame].fontsize = feedbackAttributes and feedbackAttributes.fontsize or FEEDBACK_FONTSIZE
		lib.frames[feedbackFrame].fontoutline = feedbackAttributes and feedbackAttributes.fontoutline or FEEDBACK_FONTOUTLINE
		lib.frames[feedbackFrame].animation = feedbackAttributes and feedbackAttributes.animation or FEEDBACK_ANIMATION
		lib.frames[feedbackFrame].direction = feedbackAttributes and feedbackAttributes.direction or FEEDBACK_DIRECTION
		lib.frames[feedbackFrame].delay = feedbackAttributes and feedbackAttributes.delay or FEEDBACK_DELAY
		lib.frames[feedbackFrame].steps = feedbackAttributes and feedbackAttributes.steps or FEEDBACK_STEPS
		lib.frames[feedbackFrame].stick = feedbackAttributes and feedbackAttributes.stick or FEEDBACK_STICK
		lib.frames[feedbackFrame]:SetParent(lib.frames[feedbackFrame].parent)
		lib.frames[feedbackFrame]:SetFrameStrata(lib.frames[feedbackFrame].strata)
		lib.frames[feedbackFrame]:SetWidth(lib.frames[feedbackFrame].width)
		lib.frames[feedbackFrame]:SetHeight(lib.frames[feedbackFrame].height)
		lib.frames[feedbackFrame]:SetPoint("CENTER", lib.frames[feedbackFrame].parent, "CENTER", lib.frames[feedbackFrame].xoffset, lib.frames[feedbackFrame].yoffset)
		lib.frames[feedbackFrame]:SetBackdrop({ bgFile = "Interface/Tooltips/UI-Tooltip-Background", edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 16, tileSize = 16, tile = true, insets = { left = 4, right = 4, top = 4, bottom = 4 } })
		lib:Lock(feedbackFrame)
	end
end

--[[
Notes:
	Delete a feedback frame
Arguments:
	string - the frame name
Example:
	local FB = LibStub("LibFeedback-0.1")
	FB:Create("MyFeedback")
	FB:Destroy("MyFeedback")
]]
function lib:Destroy(feedbackFrame)
	if feedbackFrame and type(feedbackFrame) == 'string' and lib.frames[feedbackFrame] then
		lib.objects[feedbackFrame] = nil
		lib.frames[feedbackFrame] = nil
	end
end

--[[
Notes:
	Add data to a feedback frame
Arguments:
	string - the frame name
	table - the feedback data
		table.text string - the text message
		table.icon string [optional] - the icon to use if enabled [default = ""]
		table.sticky boolean [optional] - produce sticky message if enabled [default=false]
Example:
	local FB = LibStub("LibFeedback-0.1")
	FB:Create("MyFeedback")
	FB:AddData("MyFeedback", { text = "Hello, World of Warcraft!" })
]]
function lib:AddData(feedbackFrame, feedbackObject)
	if feedbackFrame and type(feedbackFrame) == 'string' and lib.frames[feedbackFrame] and feedbackObject and type(feedbackObject) == 'table' and feedbackObject.text and type(feedbackObject.text) == 'string' then
		if lib.frames[feedbackFrame].objects and lib.frames[feedbackFrame].objects < 64 then
			lib.frames[feedbackFrame].objects = lib.frames[feedbackFrame].objects + 1
		else
			lib.frames[feedbackFrame].objects = 1
		end
		if not lib.objects[feedbackFrame] then
			lib.objects[feedbackFrame] = {}
		end
		feedbackObject.step = 0
		feedbackObject.xoffset = 0
		feedbackObject.yoffset = 0
		feedbackObject.number = lib.frames[feedbackFrame].objects
		feedbackObject.fonttype = feedbackObject.fonttype or lib.frames[feedbackFrame].fonttype
		feedbackObject.fontsize = feedbackObject.fontsize or lib.frames[feedbackFrame].fontsize
		feedbackObject.fontoutline = feedbackObject.fontoutline or lib.frames[feedbackFrame].fontoutline
		feedbackObject.fontcolor = feedbackObject.fontcolor or lib.frames[feedbackFrame].fontcolor
		feedbackObject.sticky = feedbackObject.sticky or false
		tinsert(lib.objects[feedbackFrame], feedbackObject)
	end
end

--[[
Notes:
	Lock a feedback frame
Arguments:
	string - the frame name
Example:
	local FB = LibStub("LibFeedback-0.1")
	FB:Create("MyFeedback")
	FB:Lock("MyFeedback")
]]
function lib:Lock(feedbackFrame)
	if feedbackFrame and type(feedbackFrame) == 'string' and lib.frames[feedbackFrame] then
		lib.frames[feedbackFrame]:SetMovable(false)
		lib.frames[feedbackFrame]:EnableMouse(false)
		lib.frames[feedbackFrame]:SetBackdropColor(1.0, 1.0, 1.0, 0.0)
		lib.frames[feedbackFrame]:SetBackdropBorderColor(1.0, 1.0, 1.0, 0.0)
	end
end

--[[
Notes:
	Unlock a feedback frame
Arguments:
	string - the frame name
Example:
	local FB = LibStub("LibFeedback-0.1")
	FB:Create("MyFeedback")
	FB:Unlock("MyFeedback")
]]
function lib:Unlock(feedbackFrame)
	if feedbackFrame and type(feedbackFrame) == 'string' and lib.frames[feedbackFrame] then
		lib.frames[feedbackFrame]:SetMovable(true)
		lib.frames[feedbackFrame]:EnableMouse(true)
		lib.frames[feedbackFrame]:RegisterForDrag("LeftButton")
		lib.frames[feedbackFrame]:SetBackdropColor(1.0, 1.0, 1.0, 0.2)
		lib.frames[feedbackFrame]:SetBackdropBorderColor(1.0, 1.0, 1.0, 0.2)
		lib.frames[feedbackFrame]:SetScript("OnDragStart", function ()
			lib.frames[feedbackFrame]:StartMoving()
		end)
		lib.frames[feedbackFrame]:SetScript("OnDragStop", function()
			lib.frames[feedbackFrame]:StopMovingOrSizing()
			_, _, _, lib.frames[feedbackFrame].xoffset, lib.frames[feedbackFrame].yoffset = lib.frames[feedbackFrame]:GetPoint()
			self.callbacks:Fire("OnFeedbackFrameMoved", feedbackFrame, lib.frames[feedbackFrame].xoffset, lib.frames[feedbackFrame].yoffset)
		end)
	end
end
